<?php if ( ! defined('BASEPATH')) exit('No direct script access allowed');

/**
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * THIS SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS," AND COPYRIGHT
 * HOLDERS MAKE NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO, WARRANTIES OF MERCHANTABILITY OR
 * FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE
 * OR DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS,
 * COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.COPYRIGHT HOLDERS WILL NOT
 * BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL
 * DAMAGES ARISING OUT OF ANY USE OF THE SOFTWARE OR DOCUMENTATION.
 * 
 * A copy of the GNU General Public License is available at:
 * 
 * 		http://gnu.org/licenses/
 * 
 * @appname		View
 * @version		0.1.2
 * @date		06/01/2010
 * 
 * @license		GNU General Public License
 * @author		Braden Schaeffer
 * @link		http://manasto.info
 * 
 * @file		View.php
 * 
 * Copyright (c) 2010
 */

class View {

	// Empty variables
	var $views				= array();

	// Header data variables
	var $page_title 		= '';
	var $css				= array();
	var $js					= array();
	var $js_plugins 		= array();
	var $js_libraries		= array();

	// Cache variables
	var $cache_enabled  	= FALSE;
	var $cache_exists   	= FALSE;
	var $do_write_cache 	= FALSE;
	var $force_cache_off	= FALSE;
	var $cache				= "";
	var $cache_data			= array(
								'header' => array('cache' => FALSE),
								'footer' => array('cache' => FALSE)
							);

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Constructor
	-------------------------------------------------------*/

	function View()
	{
		$this->CI =& get_instance();
		
		$this->_init();
		
		log_message('debug', "View Class Initialized");
	}

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Initialize
	*
	* Configure the library
	-------------------------------------------------------*/

	private function _init()
	{
		// Get the current uri segments
		$this->uri = $this->CI->uri->segment_array();

		$config = $this->CI->config;
		
		$config->load("view", TRUE);

		// Should we enable the output profiler?
		if( $config->item("enable_output_profiler", "view") === TRUE)
		{
			$this->CI->output->enable_profiler(TRUE);
		}
		
		// --------------------------------------------------------------------------

		// Set the default title
		$this->site_title = $config->item("site_title", "view");

		// Set the page title format
		$this->page_title_format = $config->item("page_title_format", "view");
		
		// --------------------------------------------------------------------------

		// Set the URL path to the CSS files
		$this->css_path = rtrim($config->item("css_path", "view"), "/")."/";

		// Set the URL path to JS files
		$this->js_path = rtrim($config->item("js_path", "view"), "/")."/";

		// Autoload any CSS files
		$autoload_css = $config->item("autoload_css", "view");

		if( ! empty($autoload_css))
		{
			$this->add_css($autoload_css);
		}

		// Autoload any javascript files
		$autoload_js = $config->item("autoload_js", "view");

		if( ! empty($autoload_js))
		{
			$this->add_js($autoload_js);
		}
		
		// Autoload any javascript libraries
		$autoload_js_libraries = $config->item("autoload_js_libraries", "view");

		if( ! empty($autoload_js_libraries))
		{
			$this->add_js_library($autoload_js_libraries);
		}
		
		// --------------------------------------------------------------------------

		// Is caching enabled?
		$this->cache_enabled = ($config->item("cache_enabled", "view") !== TRUE) ? FALSE : TRUE;

		// Set the cache path
		$this->cache_path = rtrim($config->item("cache_path", "view"), "/")."/";

		// Set the cache expiration
		$this->cache_expiration = $config->item("cache_expiration", "view");
		
		// --------------------------------------------------------------------------
		
		// Set the hooks
		$this->header_hook = $config->item("header_hook", "view");
		$this->footer_hook = $config->item("footer_hook", "view");
	}

	// --------------------------------------------------------------------------

	/**
	 * Load
	 * 
	 * Loads the view file based on previously set data
	 *
	 * @access	public
	 * @return	void
	 */

	function load()
	{
		/*-------------------------------------------------------
		* Initialize the body of the output	
		-------------------------------------------------------*/	
	
		$this->body = '';
	
		/*-------------------------------------------------------
		* Header
		-------------------------------------------------------*/
		
		$this->_compile_header();
		
		$header = $this->_call_view_hook("header");
		
		/*-------------------------------------------------------
		* Start to build the body
		-------------------------------------------------------*/

		// Does the cache exist? If not, load the view, else, load the cache
		if( ! $this->cache_exists)
		{	
			if( ! empty($this->views))
			{
				foreach($this->views as $view)
				{
					if( ! $view['string'])
					{
						$this->body .= $this->CI->load->view($view['file'], $view['data'], TRUE);
					}
					else
					{
						$this->body .= $view['file'];
					}
				}
			}
		}
		else
		{
			$this->body = $this->cache;
		}

		/*-------------------------------------------------------
		* Footer
		-------------------------------------------------------*/
		
		$footer = $this->_call_view_hook("footer");
		
		/*-------------------------------------------------------
		* Set the Final Output
		-------------------------------------------------------*/
		
		$output = $header.$this->body.$footer;
		
		/*-------------------------------------------------------
		* Display the view
		-------------------------------------------------------*/

		// Set, then display the output
		$this->CI->output->set_output($output);
		$this->CI->output->_display();

		// If a cache exists, close the database connection and exit the app		
		if($this->cache_exists)
		{
			if(class_exists('CI_DB') AND isset($this->CI->db))
			{
				$this->CI->db->close();
			}
			
			exit();
		}

		/*-------------------------------------------------------
		* Write Cache (if necessary)
		-------------------------------------------------------*/
		
		// Are we allowed to write the cache?
		if( ! $this->cache_enabled || ! $this->do_write_cache || $this->force_cache_off)
		{
			return;
		}
		
		// Is there anything to write?
		if($this->body !== "");
		{
			$this->_write_cache($this->body);
		}
	}

	// --------------------------------------------------------------------------

	/**
	 * Add View
	 * 
	 * This function works exactly like the CI $this->load->view() method with
	 * the exact same first two parameters.
	 *
	 * @access	public
	 * @param	string	the path to the view file
	 * @param	array	the data to load into the view file
	 * @return	void
	 */

	function add_view($view_file, $view_data = array(), $string = FALSE)
	{
		$this->views[] = array(
			'file' => $view_file, 
			'data' => $view_data,
			'string' => ($string === TRUE) ? TRUE : FALSE,
		);
	}

	// --------------------------------------------------------------------------

	/**
	 * Add CSS
	 * 
	 * Add a CSS file name to the array of stylesheets to be loaded.
	 *
	 * @access	public
	 * @param	mixed	the file, or array of files, for the header to load
	 * @return	void
	 */

	function add_css($name)
	{
		// Repeat the additon if the parameter is an array
		if(is_array($name))
		{
			foreach($name as $n)
			{
				$this->add_css($n);
			}
			
			return;
		}

		// If we already added the file, don't add it again
		if(in_array($name, $this->css))
		{
			return;		
		}

		$this->css[] = $this->_parse_asset($name, "css");
	}

	// --------------------------------------------------------------------------

	/**
	 * Add JS
	 * 
	 * Add a Javascript file name to the array of scripts to be loaded.
	 *
	 * @access	public
	 * @param	mixed	the file, or array of files, for the header to load
	 * @return	void
	 */

	function add_js($name)
	{
		// Repeat the additon if the parameter is an array
		if(is_array($name))
		{
			foreach($name as $n)
			{
				$this->add_js($n);
			}
			
			return;
		}

		// If we already added the file, don't add it again
		if(in_array($name, $this->js)) return;

		$this->js[] = $this->_parse_asset($name, "js");
	}

	// --------------------------------------------------------------------------

	/**
	 * Add Javascript Plugins
	 * 
	 * Creates a javascript plugins array that will be loaded before
	 * any regular files added using the add_js method
	 *
	 * @access	public
	 * @param	mixed	the file, or array of files, for the header to load
	 * @return	void
	 */

	function add_js_plugin($name)
	{
		// Repeat the additon if the parameter is an array
		if(is_array($name))
		{
			foreach($name as $n)
			{
				$this->add_js_plugin($n);
			}
			
			return;
		}

		// If we already added the file, don't add it again
		if(in_array($name, $this->js_plugins)) return;

		$this->js_plugins[] = $this->_parse_asset($name, "js");
	}

	// --------------------------------------------------------------------------

	/**
	 * Add Javascript Libraries
	 * 
	 * Creates a javascript libraries array that will be loaded before
	 * any scripts added via the add_js or add_js_plugin methods
	 *
	 * @access	public
	 * @param	mixed	the file, or array of files, for the header to load
	 * @return	void
	 */

	function add_js_library($name)
	{
		// Repeat the additon if the parameter is an array
		if(is_array($name))
		{
			foreach($name as $n)
			{
				$this->add_js_library($n);
			}
			
			return;
		}

		// If we already added the file, don't add it again
		if(in_array($name, $this->js_libraries)) return;

		$this->js_libraries[] = $this->_parse_asset($name, "js");
	}

	// --------------------------------------------------------------------------

	/**
	 * Set Page Title
	 * 
	 * Sets the page_slug variable as an extra to the site_title. The two are not
	 * the same
	 *
	 * @access	public
	 * @return	void
	 */

	function set_page_title($title = "")
	{
		$this->page_title = $title;
	}

	// --------------------------------------------------------------------------

	/**
	 * Cache On
	 * 
	 * Turn caching on
	 *
	 * @access	public
	 * @param	int 	the expiration 
	 * @return	void
	 */
	
	function cache_on($expiration = 0)
	{
		if( ! $this->cache_enabled) return;
	
		$this->do_write_cache = TRUE;

		// Are we setting a new expiration?
		if(is_int($expiration) && $expiration > 0)
		{
			$this->cache_expiration = $expiration;
		}
	}

	// --------------------------------------------------------------------------

	/**
	 * Cache Off
	 * 
	 * Turn caching off
	 *
	 * @access	public
	 * @return	void
	 */

	function cache_off()
	{
		$this->do_write_cache = FALSE;
	}

	// --------------------------------------------------------------------------

	/**
	 * Force Cache Off
	 * 
	 * Disable caching, no matter what. Cannot be undone once called.
	 *
	 * @access	public
	 * @return	void
	 */

	function force_cache_off()
	{
		$this->force_cache_off = TRUE;
	}
	
	// --------------------------------------------------------------------------
	
	/*-------------------------------------------------------
	* Write Cache
	*
	* A function for writing caches for view files
	-------------------------------------------------------*/
	
	private function _write_cache($output)
	{
		$uri = $this->_get_cache_uri();

		// Set the cache expiration time in minutes
		$expires = time() + ($this->cache_expiration * 60);

		// Serialize the header data
		$cache_data = serialize($this->cache_data);

		// Set the final cached output		
		$output = $expires."<--SEP-->".$cache_data."<--SEP-->".$output;

		if( ! is_dir($this->cache_path) OR ! is_really_writable($this->cache_path))
		{
			log_message('error', "View--> Cache path is not writeable: ".$this->cache_path);
			return;
		}

		$path = $this->cache_path.$uri;

		if( ! $fp = @fopen($path, FOPEN_WRITE_CREATE_DESTRUCTIVE))
		{
			log_message('error', "View--> Unable to write cache file.");
			return;
		}

		if(flock($fp, LOCK_EX))
		{
			fwrite($fp, $output);
			flock($fp, LOCK_UN);
		}
		else
		{
			log_message('error', "View--> Unable to secure a file lock for file at: ".$path);
			return;
		}

		fclose($fp);
		@chmod($path, DIR_WRITE_MODE);

		log_message('debug', "View--> New cache created: ".$uri);	
	}

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Display Cache
	*
	* Function for handling previously cached views
	-------------------------------------------------------*/

	function _display_cache()
	{
		if( ! is_dir($this->cache_path) OR ! is_really_writable($this->cache_path))
		{
			return FALSE;
		}

		$uri = $this->_get_cache_uri();
				
		$filepath = $this->cache_path.$uri;

		// Does a cache file even exists
		if( ! @file_exists($filepath))
		{
			return FALSE;
		}

		// Can we read the cache file?
		if( ! $fp = @fopen($filepath, FOPEN_READ))
		{
			return FALSE;
		}

		flock($fp, LOCK_SH);

		$file_content = '';

		if(filesize($filepath) > 0)
		{
			$file_content = fread($fp, filesize($filepath));
		}

		// Explode the cache contents to get the header_data and expiration
		list($timestamp, $cache_data, $cache) = explode("<--SEP-->", $file_content);

		$cache_data = unserialize($cache_data);
		
		flock($fp, LOCK_UN);
		fclose($fp);

		// Has the file expired? If so we'll delete it.
		if(time() >= $timestamp)
		{
			@unlink($filepath);
			
			log_message('debug', "View--> Cache expired on ".date("Y-m-d H:i:s", $timestamp).". File deleted.");
			
			return FALSE;
		}

		// Display the cache
		$this->_load_cache($cache, $cache_data);

		log_message('debug', "View--> Cache file loaded: ".$uri);

		return TRUE;
	}

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Get Cache URI
	*
	* Called by the caching functions to create the name
	* to be used for the cache_file
	-------------------------------------------------------*/

	private function _get_cache_uri($unique_id = "")
	{
		$cahce_uri = array();

		// The Base uri string
		$cache_uri[1] = ( ! isset($this->uri[1])) ? "home" : $this->uri[1];
		$cache_uri[2] = ( ! isset($this->uri[2])) ? "index" : $this->uri[2];

		// Is there a third Portion of the URI?
		if(isset($this->uri[3]))
		{
			$cache_uri[3] = $this->uri[3];
		}

		$uri = implode("_", $cache_uri);

		return $uri;
	}

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Load Cache
	*
	* Called by the post_controller_contructor hook
	-------------------------------------------------------*/

	private function _load_cache($cache, $data)
	{
		$this->cache_exists = TRUE;
		$this->cache = $cache;
		$this->_set_cache_data("", $data);
		$this->load();
	}

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Set Cache Data
	*
	* Sets the cache data array for either the header, the
	* footer, or both if $which is equal to ""
	-------------------------------------------------------*/

	private function _set_cache_data($which = "", $data)
	{
		if(empty($which))
		{
			$this->cache_data = $data;
			return;
		}
		
		foreach($data as $k => $v)
		{
			$this->cache_data[$which][$k] = $v; 
		}
	}
	
	// --------------------------------------------------------------------------
	
	/*-------------------------------------------------------
	* Call View Hook
	*
	* Returns the result of the custom user hooks.
	-------------------------------------------------------*/

	private function _call_view_hook($location)
	{
		// If this view is part of the cache, there is no need to continue
		if($this->cache_exists && $this->cache_data[$location]['cache'] === TRUE)
		{
			return "";
		}
		
		// Get the hook configuration
		$which = $location."_hook";
		$hook = $this->$which;
		
		// Is there any hook data to worry about? If not, "we're done"
		if(empty($hook))
		{
			return '';
		}
		
		// If the hook is not an array, it's simply the location of the view file,
		// so let's load that as a string and return it
		
		if( ! is_array($hook))
		{	
			return $this->CI->load->view($hook, $this->cache_data[$location], TRUE);
		}
		
		// If a `view_file` index was set, we're dealing with a pre-configured
		// array and there is no need to run a hook
		
		if(isset($hook['view']))
		{
			return $this->_compile_hook_data($location, $hook);
		}

		// RUN THE HOOK
		
		/* Below is extracted (almost)entirely from CI's Hooks Class */

		// -----------------------------------
		// Set file path
		// -----------------------------------
		
		
		if ( ! isset($hook['filepath']) OR ! isset($hook['filename']))
		{
			log_message("error", "View--> Either `filepath` or `filename` was not configured correctly for the {$which}");
			return "";
		}
		
		$filepath = APPPATH.$hook['filepath'].'/'.$hook['filename'];

		if ( ! file_exists($filepath))
		{
			log_message("error", "View--> The file specified for the {$which} does not exist: ".$hook['filepath']);
			return "";
		}

		// -----------------------------------
		// Set class/function name
		// -----------------------------------

		$class		= FALSE;
		$function	= FALSE;
		$view_data	= "";

		if (isset($hook['class']) AND $hook['class'] != '')
		{
			$class = $hook['class'];
		}

		if (isset($hook['function']))
		{
			$function = $hook['function'];
		}

		if (isset($hook['params']))
		{
			$params = $hook['params'];
		}

		if ($class === FALSE AND $function === FALSE)
		{
			log_message("error", "View--> No Class or function was specified for the {$which}");
			return "";
		}

		// -----------------------------------
		// Call the requested method
		// -----------------------------------

		if ($class !== FALSE)
		{
			if ( ! class_exists($class))
			{
				require($filepath);
			}

			$HOOK = new $class;
			$hook_data = $HOOK->$function($this->cache_exists);
		}
		else
		{
			if ( ! function_exists($function))
			{
				require($filepath);
			}

			$hook_data = $function($this->cache_exists);
		}
		
		// --------------------------------------------------------------------------
		
		// If the hook_data is empty, there's nothing to do
		
		if(empty($hook_data))
		{
			return '';
		}
		
		// If the hook_data is not an array, it's simply the location of the view file,
		// so let's load that as a string and return it
		
		if( ! is_array($hook_data))
		{
			return $this->CI->load->view($hook_data, $this->cache_data[$location], TRUE);
		}
		
		// If no `view_file` index was passed back, something went wrong
		if( ! isset($hook_data['view']))
		{
			log_message("error", "View--> The data returned from your {$which} was not formatted correctly");
			return "";
		}
		
		// If we made it this far, let's compile the hook data and return the view
		return $this->_compile_hook_data($location, $hook_data);
	}
	
	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Compile Hook Data
	*
	* Run through the returned (or set) hook data array
	-------------------------------------------------------*/

	private function _compile_hook_data($location, $data)
	{
		// Set the view location and unset it from the data to be sent to the view
		$view = $data['view'];
		unset($data['view']);
		
		// Was there a cache index passsed back to us?
		$data['cache'] = (isset($data['cache']) && $data['cache'] === TRUE) ? TRUE : FALSE;
		
		// Add any extra view_data to the data to be passed to the view file
		if(count($data) > 0)
		{
			foreach($data as $k => $v)
			{
				if( ! in_array($k, array('js', 'css', 'page_title')))
				{
					$this->_set_cache_data($location, array($k => $v));
				}
			}
		}
		
		// Are we caching this view? If so, add it to the body of the
		// page and return a blank string
		
		if($data['cache'] === TRUE)
		{
			$this->body .= $this->CI->load->view($view, $this->cache_data[$location], TRUE);
			return '';
		}
		
		// If we made it this far, return the requested view
		return $this->CI->load->view($view, $this->cache_data[$location], TRUE);
	}

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Get Header
	*
	* Returns the loaded header view file as a string
	-------------------------------------------------------*/

	private function _compile_header()
	{
		// If we're loading a cache, we're done
		if($this->cache_exists) return array();
		
		// --------------------------------------------------------------------------
	
		// Set the title
		$header['page_title'] = $this->_format_page_title();
		
		// Set the header css array
		$header['css'] = $this->css;
		
		// Set the header javascript array
		$header['js'] = $this->js;
		
		// Are there any javascript plugins to load?
		// If so, load them before the regular scripts
		
		if(count($this->js_plugins) > 0)
		{
			$header['js'] = array_merge($this->js_plugins, $header['js']);
		}
		
		// Are there any javascript libraries to load?
		// If so, load them before everything
		
		if(count($header['js']) > 0 && count($this->js_libraries) > 0)
		{
			$header['js'] = array_merge($this->js_libraries, $header['js']);
		}
		
		$this->_set_cache_data("header", $header);
	}
	
	// --------------------------------------------------------------------------
	
	/*-------------------------------------------------------
	* Format Page Title
	*
	* Return a string to be used as the page title based
	* on the $page_title_format set in the config file
	-------------------------------------------------------*/

	private function _format_page_title()
	{
		// If there is no format, just return the site title
		if(empty($this->page_title_format) || empty($this->page_title))
		{
			return $this->site_title;
		}

		// If there are no place_holders, just return the site title
		if( ! preg_match("/(%(?:title|slug)%)/", $this->page_title_format))
		{
			return $this->site_title;
		}

		// Replace the place holders in $page_title_format
		// and return the formatted title

		$search 	= array("%title%", 			"%slug%");
		$replace	= array($this->site_title,	$this->page_title);
		
		$page_title = str_replace($search, $replace, $this->page_title_format);

		return $page_title;
	}

	// --------------------------------------------------------------------------

	/*-------------------------------------------------------
	* Parse Asset
	*
	* Adds the path and the extension to the passed asset
	* then returns it.
	-------------------------------------------------------*/

	private function _parse_asset($file, $type)
	{
		$path_name = $type."_path";
		$path = $this->$path_name;
		
		// Was the asset passed as a direct link (i.e. http://somesite.com/style.css)?
		// If not, we need to add the path set in the config file
		
		if( ! preg_match("/^https?\:\/\//", $file))
		{
			$file = $path.$file;
		}
		
		// Do we need to add the file extension?
		$match = "/\.".$type."$/i";
		
		if( ! preg_match($match, $file))
		{
			$file .= ".".$type;
		}
		
		return $file;
	}
}

/* End of file View.php */
/* Location: ./application/libraries/View.php */